#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import pytest
import pandas as pd
import numpy as np
from nzpyida.base import IdaDataBase
from nzpyida.analytics.exploration.relation_identification import *
from nzpyida.analytics.tests.conftest import TAB_NAME_TRAIN

TEST_TAB_NAME = "TEST_TAB_NAME"

@pytest.fixture(scope='module')
def idf(idadb):
    df_train = pd.DataFrame.from_dict(
    {
        "ID": range(1000),
        "A": range(1,1001),
        "B": range(1000, 0, -1),
        "C": ['p','n']*500,
        "D": [1, 0]*500,
        "E": ['a']*500 + ['b']*500,
        "F": map(lambda l: l**2, range(1000))
    }
    )
    idf = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True)
    yield idf
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)
    
@pytest.fixture(scope='function')
def clean_up(idadb: IdaDataBase):
    if idadb.exists_table_or_view(TEST_TAB_NAME):
        idadb.drop_table(TEST_TAB_NAME)
    yield
    if idadb.exists_table_or_view(TEST_TAB_NAME):
        idadb.drop_table(TEST_TAB_NAME)

class TestCorr:
    def test_strong_correlation(self, idf, clean_up):
        out_df = corr(idf, in_column=["A", "B"], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CORRELATION', "C"])
        assert len(out_df) == 2
        assert all(out_df["CORRELATION"].head().values < -0.99)

    def test_weak_correlation(self, idf, clean_up):
        out_df = corr(idf, in_column=["A", "D"], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CORRELATION'])
        assert len(out_df) == 1
        assert 0.01 > out_df.head()[0] > -0.01
    
    def test_no_correlation(self, idf, clean_up):
        out_df = corr(idf, in_column=["A", "D"], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CORRELATION', "C"])
        assert len(out_df) == 2
        assert all(out_df["CORRELATION"].head().values == [0, 0])

class TestCov:
    def test_strong_covariance(self, idf, clean_up):
        out_df = cov(idf, in_column=['ID', 'A'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE'])
        assert len(out_df) == 1
        max_cov = np.cov(range(1, 1001),range(1, 1001))[0][0]
        assert all(0.99*max_cov < out_df["COVARIANCE"].head().values < 1.01*max_cov)
    
    def test_weak_covariance(self, idf, clean_up):
        out_df = cov(idf, in_column=['A', 'D'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE'])
        assert len(out_df) == 1
        assert np.abs(out_df["COVARIANCE"].head().values) < 0.5


    def test_no_covariance(self, idf, clean_up):
        out_df = cov(idf, in_column=['A', 'D'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE', "C"])
        assert len(out_df) == 2
        assert all( out_df["COVARIANCE"].head().values == [0, 0])

class TestMutualInfo:
    def test_strong_corelation(self, idf, clean_up):
        out_df = mutual_info(idf, ['A', 'B'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['MUTUALINFO', 'OVER_C'])
        assert len(out_df) == 2
        assert all(out_df["MUTUALINFO"].head().values < 10)
    
    @pytest.mark.skip # "Attribute ... must be GROUPed or used in an aggregate function" error
    def test_weak_correlation(self, idf, clean_up):
        out_df = mutual_info(idf, ['A', 'D'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['MUTUALINFO', 'OVER_C'])
        assert len(out_df) == 1

    @pytest.mark.skip # "Attribute ... must be GROUPed or used in an aggregate function" error
    def test_no_correlation(self, idf, clean_up):
        out_df = mutual_info(idf, ['A', 'D'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['MUTUALINFO', 'OVER_C'])
        assert len(out_df) == 2

# TODO: poorly readable output
class TestCovarianceMatrix:
    def test_strong_corelation(self, idf, clean_up):
        out_df = covariance_matrix(idf, ['A', 'B'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE', 'C'])
        assert len(out_df) == 2
    
    def test_weak_correlation(self, idf, clean_up):
        out_df = covariance_matrix(idf, ['A', 'D'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE'])
        assert len(out_df) == 1

    def test_no_correlation(self, idf, clean_up):
        out_df = covariance_matrix(idf, ['A', 'D'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['COVARIANCE', 'C'])
        assert len(out_df) == 2

class TestSpearmanCorr:
    def test_strong_corelation(self, idf, clean_up):
        out_df = spearman_corr(idf, ['A', 'B'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['C', 'RHO', 'N'])
        assert len(out_df) == 2
        assert all( out_df["RHO"].head().values == -1)
    
    def test_weak_correlation(self, idf, clean_up):
        out_df = spearman_corr(idf, ['A', 'D'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['SPEARMAN_CORR'])
        assert len(out_df) == 1
        assert all(abs(out_df["SPEARMAN_CORR"].head().values) < 0.05)
        
    def test_no_correlation(self, idf, clean_up):
        out_df = spearman_corr(idf, ['A', 'D'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['C', 'RHO', 'N'])
        assert len(out_df) == 2
        assert all(out_df["RHO"].head().values == 0)

class TestChisq:
    # I believe last assertsion should be stronger (... > 0.95)
    def test_strong_correlation(self, idf, clean_up):
        out_df = chisq(idf, ['A', 'B'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'CHI2STATISTIC', 'DF', 'OVER_NO_GROUP'])
        assert len(out_df) == 1
        assert out_df['PERCENTAGE'].head().values[0] > 0.75

    @pytest.mark.skip # "Attribute ... must be GROUPed or used in an aggregate function" error
    def test_weak_correlation(self, idf, clean_up):
        out_df = chisq(idf, ['A', 'D'], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'CHI2STATISTIC', 'DF', 'OVER_NO_GROUP'])
        assert len(out_df) == 1
        assert out_df['PERCENTAGE'].head().values[0] < 0.05
    
    @pytest.mark.skip # "Attribute ... must be GROUPed or used in an aggregate function" error
    def test_no_correlation(self, idf, clean_up):
        out_df = chisq(idf, ['A', 'D'], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'CHI2STATISTIC', 'DF', 'OVER_NO_GROUP'])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values == 0)


class TestTMeTest:
    def test_linear_column_correct_value(self, idf, clean_up):
        out_df = t_me_test(idf, in_column="A", mean_value=500, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_ME_TEST'])
        assert len(out_df) == 1
        assert 0.05 < out_df['PERCENTAGE'].head().values[0] < 0.95
    
    def test_linear_column_incorrect_value(self, idf, clean_up):
        out_df = t_me_test(idf, in_column="A", mean_value=400, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_ME_TEST'])
        assert len(out_df) == 1
        assert out_df['PERCENTAGE'].head().values[0] > 0.95
    
    def test_linear_column_incorrect_value_with_by_column(self, idf, clean_up):
        out_df = t_me_test(idf, in_column="A", mean_value=550, by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_ME_TEST', "C"])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values < 0.05)
    
    def test_constant_column(self, idf, clean_up):
        out_df = t_me_test(idf, in_column="D", mean_value=1, by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_ME_TEST', "C"])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values == None)

class TestTUmdTest:
    def test_linear_column_correct_value(self, idf, clean_up):
        out_df = t_umd_test(idf, in_column="A", class_column='C:p:n', out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_UMD_TEST'])
        assert len(out_df) == 1
        assert 0.45 < out_df['PERCENTAGE'].head().values[0] < 0.55
    
    def test_linear_column_incorrect_value(self, idf, clean_up):
        out_df = t_umd_test(idf, in_column="A", class_column='E:a:b', by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_UMD_TEST', "C"])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values < 0.05)
    
class TestTPmdTest:
    def test_good_diff(self, idf, clean_up):
        out_df = t_pmd_test(idf, in_column=["A:X", "B:Y"], expected_diff=0, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'TSTUDPAIREDMEANDIFF'])
        assert len(out_df) == 1
        assert 0.45 < out_df['PERCENTAGE'].head().values[0] < 0.55
    
    def test_slightly_wrong_diff(self, idf, clean_up):
        out_df = t_pmd_test(idf, in_column=["A:X", "B:Y"], expected_diff=1, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'TSTUDPAIREDMEANDIFF'])
        assert len(out_df) == 1
        assert 0.05 < out_df['PERCENTAGE'].head().values[0] < 0.95
    
    def test_wrong_diff(self, idf, clean_up):
        out_df = t_pmd_test(idf, in_column=["A:X", "B:Y"], by_column="E", expected_diff=10, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'TSTUDPAIREDMEANDIFF', "E"])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values == [0, 1])


class TestTLsTest:
    @pytest.mark.skip # TODO: this function is failing, when slope is exactly right
    def test_good_slope(self, idf, clean_up):
        out_df = t_ls_test(idf, in_column=["A:X", "B:Y"], slope=-1, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_LS_TEST'])
        assert len(out_df) == 1
        assert 0.05 < out_df['PERCENTAGE'].head().values[0] < 0.95
    
    def test_wrong_slope(self, idf, clean_up):
        out_df = t_ls_test(idf, in_column=["A:X", "B:Y"], slope=1, by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_LS_TEST', 'C'])
        assert len(out_df) == 2
        assert all(out_df['PERCENTAGE'].head().values == 0)

    @pytest.mark.skip # TODO: this function is not working correctly, as slope close to correct value is treated as very bad one
    def test_slightly_wrong_slope(self, idf, clean_up):
        out_df = t_ls_test(idf, in_column=["A:X", "B:Y"], slope=-0.99999, out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['PERCENTAGE', 'T_LS_TEST'])
        assert len(out_df) == 1
        assert 0.05 < out_df['PERCENTAGE'].head().values[0] < 0.95


class TestMwwTest:
    def test_linear_column_similar_values_with_by_column(self, idf, clean_up):
        out_df = mww_test(idf, in_column="A", class_column="C", by_column="E", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['E', 'N', 'N1', 'N2', 'USTAT0', 'U2STAT0', 'ALPHASUM', 'NORM1',
                                      'USTAT1', 'NORM2', 'U2STAT1', 'MU_U1', 'SIGMA_U', 'ZSTAT', 'PP',
                                      'USTAT', 'U2STAT', 'MU_U', 'NOGROUPS', 'MESSAGE', 'LOWER'])
        assert len(out_df) == 2
        assert all(out_df['PP'].head().values > 0.05)
        assert all(out_df['LOWER'].head().values == 'p')
    
    def test_linear_column_similar_values_without_by_column(self, idf, clean_up):
        out_df = mww_test(idf, in_column="A", class_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['USTAT', 'U2STAT', 'MUU', 'SIGMAU', 'ZSTAT', 'PP', 'LOWER'])
        assert len(out_df) == 1
        assert out_df['PP'].head().values[0] > 0.05
        assert out_df['LOWER'].head().values[0] == 'p'
    
    def test_linear_column_different_values_with_by_column(self, idf, clean_up):
        out_df = mww_test(idf, in_column="A", class_column="E", by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['C', 'N', 'N1', 'N2', 'USTAT0', 'U2STAT0', 'ALPHASUM', 'NORM1',
                                      'USTAT1', 'NORM2', 'U2STAT1', 'MU_U1', 'SIGMA_U', 'ZSTAT', 'PP',
                                      'USTAT', 'U2STAT', 'MU_U', 'NOGROUPS', 'MESSAGE', 'LOWER'])
        assert len(out_df) == 2
        assert all(out_df['PP'].head().values < 0.05)
        assert all(out_df['LOWER'].head().values == 'a')


class TestWilcoxonTest:
    # function wrogly indicates column with lower values, need to change last assertions
    def test_similar_columns(self, idf, clean_up):
        out_df = wilcoxon_test(idf, in_column=["A", "B"], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['SSTAT', 'WSTAT', 'ZSTAT', 'NOITEM', 'PP', 'LOWER'])
        assert len(out_df) == 1
        assert out_df['PP'].head().values[0] > 0.05
    
    # function wrogly indicates column with lower values, need to change last assertions
    def test_similar_columns_with_by_column(self, idf, clean_up):
        out_df = wilcoxon_test(idf, in_column=["A", "B"], by_column="E", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['E', 'N', 'SSTAT', 'WSTAT', 'ZSTAT', 'PP', 'LOWER'])
        assert len(out_df) == 2
        assert all(out_df['PP'].head().values < 0.05)
        assert out_df[out_df["E"]=='a']["LOWER"].head().values[0] == '"B"'
        assert out_df[out_df["E"]=='b']["LOWER"].head().values[0] == '"A"'
    
    # function wrogly indicates column with lower values, need to change last assertion
    def test_different_columns(self, idf, clean_up):
        out_df = wilcoxon_test(idf, in_column=["A", "D"], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['SSTAT', 'WSTAT', 'ZSTAT', 'NOITEM', 'PP', 'LOWER'])
        assert len(out_df) == 1
        assert out_df['PP'].head().values[0] < 0.05
        assert out_df['LOWER'].head().values[0] == '"A"'

# TODO: poorly readable output
class TestCanonicalCorr:
    def test_strong_correlation(self, idf, clean_up):
        out_df = canonical_corr(idf, in_column=["A", "B"], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CANONICAL_CORRELATION'])
        assert len(out_df) == 1
    
    def test_weak_correlation(self, idf, clean_up):
        out_df = canonical_corr(idf, in_column=["A", "D"], out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CANONICAL_CORRELATION'])
        assert len(out_df) == 1
    
    def test_no_correlation(self, idf, clean_up):
        out_df = canonical_corr(idf, in_column=["A", "D"], by_column="C", out_table=TEST_TAB_NAME)
        assert out_df
        assert all(out_df.columns == ['CANONICAL_CORRELATION', "C"])
        assert len(out_df) == 2


class TestAnovaCrdTest:
    def test_simmilar_columns(self, idf, clean_up):
        out_df = anova_crd_test(idf, in_column=['D'], treatment="E", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['TOTNO', 'TOTSU', 'TOTMEAN', 'TOTSS', 'SSCTOT', 'SSCBETWEEN', 
                                      'DFBETWEEN', 'SSCWITHIN', 'DFWITHIN', 'F', 'P'])
        assert len(out_df) == 1
        assert out_df['P'].head().values[0] < 0.95 
    
    def test_dfferent_columns(self, idf, clean_up):
        out_df = anova_crd_test(idf, in_column=['A'], treatment="E", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['TOTNO', 'TOTSU', 'TOTMEAN', 'TOTSS', 'SSCTOT', 'SSCBETWEEN', 
                                      'DFBETWEEN', 'SSCWITHIN', 'DFWITHIN', 'F', 'P'])
        assert len(out_df) == 1
        assert out_df['P'].head().values[0] > 0.95
    
    def test_multiple_columns_with_by_column(self, idf, clean_up):
        out_df = anova_crd_test(idf, in_column=['A', 'B'], treatment="E", by_column='C', out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['INCOLUMN', 'C', 'TOTNO', 'TOTSU', 'TOTMEAN', 'TOTSS', 'SSCTOT', 
                                      'SSCBETWEEN', 'DFBETWEEN', 'SSCWITHIN', 'DFWITHIN', 'F', 'P'])
        assert len(out_df) == 4


class TestAnovaRbdTest:
    def test_one_column(self, idf, clean_up):
        out_df = anova_rbd_test(idf, in_column=['A'], treatment="E", block="C", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['BLSSCBETWEEN', 'BLDFBETWEEN','SSCWITHIN', 'DFWITHIN', 'FBL', 'PBL', 
                                      'GRSSCBETWEEN', 'GRDFBETWEEN', 'FGR', 'PGR'])
        assert len(out_df) == 1
        assert out_df['PGR'].head().values[0] > 0.95  
        assert out_df['PBL'].head().values[0] < 0.95 
    
    def test_multiple_columns(self, idf, clean_up):
        out_df = anova_rbd_test(idf, in_column=['A', 'B'], treatment="C", block='E', out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['INCOLUMN', 'BLSSCBETWEEN', 'BLDFBETWEEN','SSCWITHIN', 'DFWITHIN', 
                                      'FBL', 'PBL', 'GRSSCBETWEEN', 'GRDFBETWEEN', 'FGR', 'PGR'])
        assert len(out_df) == 2
        assert all(out_df['PGR'].head().values < 0.95)
        assert all(out_df['PBL'].head().values > 0.95)

class TestManovaOneWayTest:
    def test_idependent_columns(self, idf, clean_up):
        out_df = manova_one_way_test(idf, in_column=["A","D"], factor1="E", id_column="ID", 
                                     table_type="columns", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['ID_TASK', 'ID_MATRIX', 'ROW', 'COL', 'VAL', 'EXPLANATION',
                                      'BONFERRONI_CORR', 'TASK_NAME', 'VARIABLE_NAME'])
        assert len(out_df) == 21
        assert len(set(out_df['ID_MATRIX'].as_dataframe().values)) == 4
    
    def test_correlated_columns(self, idf, clean_up):
        out_df = manova_one_way_test(idf, in_column=["A","F"], factor1="C", id_column="ID", 
                                     table_type="columns", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['ID_TASK', 'ID_MATRIX', 'ROW', 'COL', 'VAL', 'EXPLANATION',
                                      'BONFERRONI_CORR', 'TASK_NAME', 'VARIABLE_NAME'])
        assert len(out_df) == 21
        assert len(set(out_df['ID_MATRIX'].as_dataframe().values)) == 4

class TestManovaTwoWayTest:
    def test_correlated_columns(self, idf, clean_up):
        out_df = manova_two_way_test(idf, in_column=["A","F"], factor1="E", factor2="C", id_column="ID", 
                                     table_type="columns", out_table="TEST_TAB_NAME")
        assert out_df
        assert all(out_df.columns == ['ID_TASK', 'ID_MATRIX', 'ROW', 'COL', 'VAL', 'EXPLANATION',
                                      'BONFERRONI_CORR', 'TASK_NAME', 'VARIABLE_NAME'])
        assert len(out_df) == 59
        assert len(set(out_df['ID_MATRIX'].as_dataframe().values)) == 8


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The naive Bayes classifier is a simpler classification algorithm than most,
which makes it quick and easy to apply. While it does not compete with more
sophisticated algorithms with respect to classification accuracy, in some
cases it may be able to deliver similar results in a fraction of the
computation time.
"""
from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.classification import Classification

class NaiveBayesClassifier(Classification):
    """
    Naive Bayes classifier
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the classifier class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """
        super().__init__(idadb, model_name)
        self.fit_proc = 'NAIVEBAYES'
        self.predict_proc = 'PREDICT_NAIVEBAYES'

    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None, 
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, disc: str=None, bins: int=10):
        """
        Builds a Naive Bayes model.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        in_columns : List[str], optional
            the input table columns with special properties, separated by a semi-colon (;).
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns. Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be detected
            automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight' is unsupported,
            i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        disc : str, optional
            discretization type for numeric columns [ew, ef, em]

        bins : int, optional
            default number of bins for numeric columns
        """

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'disc': disc,
            'bins': bins
        }

        self._fit(in_df=in_df, params=params)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None, out_table_prob: str=None, mestimation: str=None):
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        out_table_prob : str, optional
            if specified, the probability output table where class probability predictions
            will be stored

        mestimation : str, optional
            flag indicating to use m-estimation for probabilities. This kind of estimation of
            probabilities may be slower but can give better results for small or heavy
            unbalanced datasets.

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column,
            'outtableProb': out_table_prob,
            'mestimation': mestimation
            }

        return self._predict(in_df=in_df, params=params, out_table=out_table)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
Tree-shaped Bayesian networks formally belong to the data exploration category. 
However, this algorithm is considerably more complex than other data exploration 
algorithms and not as widely known, warranting detailed description.

A Bayesian network can be considered a graphical representation of probabilistically 
described relationships within a set of attributes, allowing probabilistic inference 
to be performed. The representation is created by extracting the structural properties 
of the distribution from the data.

Creating and using general Bayesian networks are algorithmically and computationally 
complex. Tree- shaped Bayesian networks, however, constitute a simplified subclass 
of Bayesian networks with restrictions imposed on the type of attribute relationships 
that can be discovered and represented. The restrictions permit simpler and more efficient 
algorithms as well as more straightforward interpretation. Tree-shaped Bayesian networks 
may be not sufficient for highly-accurate prediction, but provide an excellent 
qualitative description of the relationship structure observed in the data
"""

from typing import Tuple, List
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, materialize_df, make_temp_table_name
from nzpyida.analytics.utils import get_auto_delete_context
from nzpyida.analytics.predictive.regression import Regression
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling


class TreeBayesNetwork(Regression):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Tree Shaped Bayesian Network class

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "TBNET_GROW"
        self.predict_proc = "TBNET_APPLY"

    def fit(self, in_df: IdaDataFrame, in_columns: List[str]=None, base_index: int=777,
            sample_size: int=None, talk: str=None, size_warning: str=None, edge_lab_sort: str=None, col_def_type: str=None, 
            col_def_role: str=None, col_properties_table: str=None) -> None:
        """
        Builds a tree-like Bayesian Network for continuous variables. A spanning tree is 
        constructed joining all the variables on grounds of most strong correlations. 
        This gives the user an overview of most significant interrelations governing 
        the whole set of variables

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame
        
        in_columns : List[str]
            List of the input dataframe  columns with special properties. 
            Each column is followed by one or several of the following properties:
            - type: ':nom' (for nominal), ':cont' (for continuous). By default, 
              all numerical types are continuous, other types are nominal
            - role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').

            If the parameter is undefined, all columns of the input dataframe have default properties. 
            Note that this procedure only accepts continuous columns with role 'input'
        
        base_index : int, optional
            the numeric id to be assigned to the first variable
        
        sample_size : int, optional
            the sample size to take if the number of records is too large
        
        talk : str, optional
            if talk=yes then additional information on progress will be displayed
        
        size_warning : str, optional
            if sizewarn=yes then no exception is thrown when there are less records than 
            3 times the number of columns. Instead, a notice is displayed and the stored 
            procedure returns 'sizewarn'
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        
        col_def_type : str, optional
            default type of the input dataaframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input dataframe columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input dataframe columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')
        """
        params = {
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colPropertiesTable': col_properties_table,
            'baseidx': base_index,
            'samplesize': sample_size,
            'talk': talk,
            'edgelabsort': edge_lab_sort
        }
        if self.fit_proc == 'TBNET_GROW':
            params['sizewarn'] = size_warning

        self._fit(in_df=in_df, params=params, needs_id=False)
    
    def predict(self, in_df: IdaDataFrame, target_column: str=None, id_column: str=None,
                prediction_type: str='best', out_table: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame
        
        target_column : str
            The model variable to be predicted
        
        id_column : str, optional
            The column of the input dataframe that identifies a unique instance ID
        
        prediction_type : str, optional
            The type of prediction to be made. Valid values are best (most correlated neighbor), 
            neighbors (weighted prediction of neighbors), and nn-neighbors (non null neighbors)
        
        out_table : str, optional
            The name of the output dataframe where the predictions are to be stored
        
        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """
        params = {
            'id': id_column,
            'target': target_column,
            'type': prediction_type
        }
        return self._predict(in_df, params, out_table)
    

    def score(self, in_df: IdaDataFrame, target_column: str=None, id_column: str=None,
              prediction_type: str='best') -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input dataframe column representing the class

        id_column : str
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id
        
        prediction_type : str, optional
            The type of prediction to be made. Valid values are best (most correlated neighbor), 
            neighbors (weighted prediction of neighbors), and nn-neighbors (non null neighbors)

        Returns
        -------
        float
            the model score
        """
        params = {
            'id': id_column,
            'type': prediction_type,
            'target': target_column
        }
        self.target_column_in_output = f"{target_column}_PRED"
        return self._score(in_df, params, target_column)



class BinaryTreeBayesNetwork(TreeBayesNetwork):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Binary Tree Shaped Bayesian Network class
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "BTBNET_GROW"



class TreeAgumentedNetwork(TreeBayesNetwork):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Tree-shaped Agumented Network object
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "TANET_GROW"
        self.predict_proc = "TANET_APPLY"
    
    def fit(self, in_df: IdaDataFrame, in_model: str, class_column: str,
            edge_lab_sort: str=None) -> None:
        """
        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame
        
        in_model : str
            the name of the input Bayesian Network model

        class_column : str
            the target class; this should be column with nominal variables
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        """
        params = {
            'inmodel': in_model,
            'class': class_column,
            'edge_lab_sort': edge_lab_sort
        }
        self._fit(in_df, params, needs_id=False)


class MultiTreeBayesNetwork(TreeBayesNetwork):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Multi Tree-shaped Bayesian Network object
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "MTBNET_GROW"
        self.predict_proc = "TANET_APPLY"
    
    def fit(self, in_df: IdaDataFrame, class_column: str, in_columns: List[str]=None, 
            base_index: int=None, sample_size: int=None, talk: str=None, 
            edge_lab_sort: str=None, col_def_type: str=None, col_def_role: str=None, 
            col_properties_table: str=None) -> None:
        """
        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        class_column : str
            the target class; this should be column with nominal variables

        in_columns : List[str]
            List of the input dataframe columns with special properties. 
            Each column is followed by one or several of the following properties:
            - type: ':nom' (for nominal), ':cont' (for continuous). By default, 
              all numerical types are continuous, other types are nominal
            - role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').

            If the parameter is undefined, all columns of the input table have default properties. 
            Note that this procedure only accepts continuous columns with role 'input'
            Addition-ally, each column is followed by a colon (:) and either X or Y to distinguish 
            the two sets of variables.
        
        base_index : int, optional
            the numeric id to be assigned to the first variable
        
        sample_size : int, optional
            the sample size to take if the number of records is too large
        
        talk : str, optional
            if talk=yes then additional information on progress will be displayed
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        
        col_def_type : str, optional
            default type of the input dataframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input dataframe columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input dataframe columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input dataframe column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')
        """
        params = {
            'class': class_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colPropertiesTable': col_properties_table,
            'baseidx': base_index,
            'samplesize': sample_size,
            'talk': talk,
            'edgelabsort': edge_lab_sort
        }
        self._fit(in_df, params, needs_id=False)





class TreeBayesNetworkBase(PredictiveModeling):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates base class for group of Tree Shaped Bayesian Network models
        """
        super().__init__(idadb, model_name)
    
    def _grow(self, in_df: IdaDataFrame, params: dict):
        """
        Grows the Tree Shaped Bayesian Network on input data and returns
        dataframe with statistics

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        params : dict
            the dictionary of attributes used for running procedure

        Returns
        -------
        IdaDataFrame
            the data frame containing statistics
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        ModelManager(self.idadb).drop_model(self.model_name)

        temp_view_name, need_delete = materialize_df(in_df)
        params['intable'] = temp_view_name

        params_s = map_to_props(params)

        try:
            in_df.ida_query(f'call NZA..TBNET1G(\'{params_s}\')')
        finally:
            if need_delete:
                in_df._idadb.drop_view(temp_view_name)
        
        table_name = f"INZA.nza_meta_{self.model_name}_model"
            
        if not in_df._idadb.exists_table_or_view(table_name):
            # stored procedure call was successful by did not produce a table
            return None
        
        out_df = IdaDataFrame(in_df._idadb, table_name)
        return out_df


class TreeBayesNetwork1G(TreeBayesNetworkBase):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Tree-shaped Bayesian Network 1G object
        """
        super().__init__(idadb, model_name)
        self.predict_proc = "TBNET1G"

    def grow(self, in_df: IdaDataFrame, in_columns: List[str]=None, base_index: int=777, 
            sample_size: int=330000, talk: str=None, no_check: str=None, edge_lab_sort: str=None, 
            col_def_type: str=None, col_def_role: str=None, 
            col_properties_table: str=None) -> IdaDataFrame:
        """
        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        in_columns : List[str]
            List of the input dataframe columns with special properties. 
            Each column is followed by one or several of the following properties:
            - type: ':nom' (for nominal), ':cont' (for continuous). By default, 
              all numerical types are continuous, other types are nominal
            - role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').

            If the parameter is undefined, all columns of the input table have default properties. 
            Note that this procedure only accepts continuous columns with role 'input'
            Addition-ally, each column is followed by a colon (:) and either X or Y to distinguish 
            the two sets of variables.
        
        base_index : int, optional
            the numeric id to be assigned to the first variable
        
        sample_size : int, optional
            the sample size to take if the number of records is too large
        
        talk : str, optional
            if talk=yes then additional information on progress will be displayed
        
        no_check : str, optional
            if nocheck=yes then no exception is thrown when a column in <in_columns> 
            does not exis
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        
        col_def_type : str, optional
            default type of the input dataframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input dataframe columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input dataframe columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input dataframe column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')
        
        Returns
        -------
        IdaDataFrame
            the data frame containing statistics
        """
        params = {
            'model': self.model_name,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colPropertiesTable': col_properties_table,
            'baseidx': base_index,
            'samplesize': sample_size,
            'talk': talk,
            'nocheck': no_check,
            'edgelabsort': edge_lab_sort
            }
        
        return self._grow(in_df, params)
    

class TreeBayesNetwork2G(TreeBayesNetworkBase):
    def __init__(self, idadb, model_name) -> None:
        """
        Creates Tree-shaped Bayesian Network 2G object
        """
        super().__init__(idadb, model_name)
        self.predict_proc = "TBNET2G"

    def grow(self, in_df: IdaDataFrame, in_columns: List[str]=None, base_index: int=777, 
            talk: str=None, no_check: str=None, edge_lab_sort: str=None, col_def_type: str=None, 
            col_def_role: str=None, col_properties_table: str=None) -> IdaDataFrame:
        """
        Builds a tree-like Bayesian Network for continuous variables.
        A spanning tree is constructed joining all the variables on grounds of most strong 
        correlations. This gives the user an overview of most significant interrelations 
        governing the whole set of variables.

        The stored procedure operates with two sets of variables and the resulting tree 
        will be bi-partite. The correlations between variables within each set will not 
        be calculated. This feature is useful when the two sets characterize distinct 
        objects and only links between the objects are of interest
        
        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        in_columns : List[str]
            List of the input dataframe columns with special properties. 
            Each column is followed by one or several of the following properties:
            - type: ':nom' (for nominal), ':cont' (for continuous). By default, 
              all numerical types are continuous, other types are nominal
            - role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').

            If the parameter is undefined, all columns of the input table have default properties. 
            Note that this procedure only accepts continuous columns with role 'input'
            Addition-ally, each column is followed by a colon (:) and either X or Y to distinguish 
            the two sets of variables.
        
        base_index : int, optional
            the numeric id to be assigned to the first variable
        
        talk : str, optional
            if talk=yes then additional information on progress will be displayed
        
        no_check : str, optional
            if nocheck=yes then no exception is thrown when a column in <in_columns> 
            does not exis
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        
        col_def_type : str, optional
            default type of the input dataframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input dataframe columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input dataframe columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input dataframe column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        Returns
        -------
        IdaDataFrame
            the data frame containing statistics
        """
        params = {
            'model': self.model_name,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colPropertiesTable': col_properties_table,
            'baseidx': base_index,
            'talk': talk,
            'nocheck': no_check,
            'edgelabsort': edge_lab_sort
            }
        
        return self._grow(in_df, params)

class TreeBayesNetwork1G2P(TreeBayesNetworkBase):
    def __init__(self, idadb, model_name) -> None:
        """
        Creates Tree-shaped Bayesian Network 1G2P object
        """
        super().__init__(idadb, model_name)
        self.predict_proc = "TBNET1G2P"
    
    def grow(self, in_df: IdaDataFrame, in_columns: List[str]=None, base_index: int=777, 
            talk: str=None, no_check: str=None, edge_lab_sort: str=None, col_def_type: str=None, 
            col_def_role: str=None, col_properties_table: str=None) -> IdaDataFrame:
        """
        This stored procedure builds a tree-like Bayesian Network for continuous variables. 
        A spanning tree is constructed joining all the variables on grounds of most strong 
        correlations. This gives the user an overview of most significant interrelations 
        governing the whole set of variables.

        The stored procedure constructs the tree in an incremental manner. It calculates 
        correlations on one set of variables, then on the other set of variables, then 
        between variables of the 2 sets. The final model is obtained by joining the 
        three sub-models
        
        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        in_columns : List[str]
            List of the input dataframe columns with special properties. 
            Each column is followed by one or several of the following properties:
            - type: ':nom' (for nominal), ':cont' (for continuous). By default, 
              all numerical types are continuous, other types are nominal
            - role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').

            If the parameter is undefined, all columns of the input table have default properties. 
            Note that this procedure only accepts continuous columns with role 'input'
            Addition-ally, each column is followed by a colon (:) and either X or Y to distinguish 
            the two sets of variables.
        
        base_index : int, optional
            the numeric id to be assigned to the first variable
        
        talk : str, optional
            if talk=yes then additional information on progress will be displayed
        
        no_check : str, optional
            if nocheck=yes then no exception is thrown when a column in <in_columns> 
            does not exis
        
        edge_lab_sort : str, optional
            if edge_lab_sort=yes then the left end of the edge will have a name lower 
            in alphabetic order than the right one
        
        col_def_type : str, optional
            default type of the input dataframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input dataframe columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input dataframe columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input dataframe column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        Returns
        -------
        IdaDataFrame
            the data frame containing statistics
        """
        params = {
            'model': self.model_name,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colPropertiesTable': col_properties_table,
            'baseidx': base_index,
            'talk': talk,
            'nocheck': no_check,
            'edgelabsort': edge_lab_sort
            }
        
        return self._grow(in_df, params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.bayesian_networks import TreeBayesNetwork, TreeBayesNetwork1G, \
     TreeBayesNetwork1G2P, TreeBayesNetwork2G, TreeAgumentedNetwork, BinaryTreeBayesNetwork, \
     MultiTreeBayesNetwork
from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_train_reg, df_test_reg
import pytest

MOD_NAME2 = "TEST_MOD2"
OUT_TABLE_PRED2 = "OUT_TABLE_PRED2"

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='function')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if mm.model_exists(MOD_NAME2):
        mm.drop_model(MOD_NAME2)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_PRED2):
        idadb.drop_table(OUT_TABLE_PRED2)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if mm.model_exists(MOD_NAME2):
        mm.drop_model(MOD_NAME2)
    if idadb.exists_table(OUT_TABLE_PRED2):
        idadb.drop_table(OUT_TABLE_PRED2)

@pytest.fixture(scope='module')
def idf_train_nom(idadb: IdaDataBase):
    TAB_NAME_TRAIN_NOM = TAB_NAME_TRAIN + "_NOM"
    df_train_reg["C"] = ['n', 'p', 'n', 'p', 'n', 'p', 'n', 'p', 'n', 'p']*100
    yield idadb.as_idadataframe(df_train_reg, tablename=TAB_NAME_TRAIN_NOM, clear_existing=True)
    if idadb.exists_table(TAB_NAME_TRAIN_NOM):
        idadb.drop_table(TAB_NAME_TRAIN_NOM)

@pytest.fixture(scope='module')
def idf_test_nom(idadb: IdaDataBase):
    TAB_NAME_TEST_NOM = TAB_NAME_TEST + "_NOM"
    df_test_reg["C"]=['n', 'n', 'p', 'p', 'n']
    yield idadb.as_idadataframe(df_test_reg, tablename=TAB_NAME_TEST_NOM, clear_existing=True)
    if idadb.exists_table(TAB_NAME_TEST_NOM):
        idadb.drop_table(TAB_NAME_TEST_NOM)

@pytest.fixture(scope='module')
def idf_train(idadb: IdaDataBase):
    yield idadb.as_idadataframe(df_train_reg, tablename=TAB_NAME_TRAIN, clear_existing=True)
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)

@pytest.fixture(scope='module')
def idf_test(idadb: IdaDataBase):
    yield idadb.as_idadataframe(df_test_reg, tablename=TAB_NAME_TEST, clear_existing=True)
    if idadb.exists_table(TAB_NAME_TEST):
        idadb.drop_table(TAB_NAME_TEST)

def test_tree_bayes_network(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame,
                            idf_test: IdaDataFrame, clear_up):
    model = TreeBayesNetwork(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)

    model.fit(idf_train, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME)

    pred  = model.predict(idf_test, target_column="B", id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ["ID", "B_PRED"])
    # assert any(round(x) == y for x, y in zip(list(pred.as_dataframe()['B_PRED'].values),  [2, 4, 2223, -999, 11112]))


def test_binary_tree_bayes_network(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame ,
                                   idf_test: IdaDataFrame , clear_up):
    model = BinaryTreeBayesNetwork(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)

    model.fit(idf_train, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME)

    pred  = model.predict(idf_test, target_column="B", id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ["ID", "B_PRED"])
    assert any(round(x) == y for x, y in zip(list(pred.head()['B_PRED'].values),  [2, 4, 2223, -999, 11112]))


def test_multi_tree_bayes_network(idadb: IdaDataBase, mm: ModelManager, idf_train_nom: IdaDataFrame ,
                                   idf_test_nom: IdaDataFrame , clear_up):
    model = MultiTreeBayesNetwork(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)

    model.fit(idf_train_nom, class_column= "C", in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME)

    pred  = model.predict(idf_test_nom, target_column="B", id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ["ID", "B_PRED"])
    assert any(round(x) == y for x, y in zip(list(pred.head()['B_PRED'].values),  [2, 4, 2223, -999, 11112]))


def test_tree_bayes_network_1g(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame,
                               clear_up):
    
    model = TreeBayesNetwork1G(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)

    outtab = model.grow(idf_train, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME)
    assert outtab
    # TODO: check output of an outtab


def test_tree_bayes_network_2g(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame,
                               clear_up):
    
    model = TreeBayesNetwork2G(idadb, MOD_NAME2)
    assert model
    assert not mm.model_exists(MOD_NAME2)

    outtab = model.grow(idf_train, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME2)
    assert outtab
    # TODO: check output of an outtab

def test_tree_bayes_network_1g2p(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame,
                               clear_up):
    
    model = TreeBayesNetwork1G2P(idadb, MOD_NAME2)
    assert model
    assert not mm.model_exists(MOD_NAME2)

    outtab = model.grow(idf_train, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME2)
    assert outtab
    # TODO: check output of an outtab


def test_tree_agumented_network(idadb: IdaDataBase, mm: ModelManager, idf_train_nom: IdaDataFrame ,
                                   idf_test_nom: IdaDataFrame , clear_up):
    model = BinaryTreeBayesNetwork(idadb, MOD_NAME2)
    assert model
    assert not mm.model_exists(MOD_NAME2)

    model.fit(idf_train_nom, in_columns=["A", "B"])
    assert mm.model_exists(MOD_NAME2)

    model2 = TreeAgumentedNetwork(idadb, MOD_NAME)
    model2.fit(idf_train_nom, in_model=MOD_NAME2, class_column="C")

    assert mm.model_exists(MOD_NAME)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from nzpyida.base import IdaDataBase
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.utils import map_to_props, call_proc_df_in_out
from typing import List

class ModelManager:
    """
    Manages in-database models.
    """

    def __init__(self, idadb: IdaDataBase):
        """
        Creates the context.

        Parameters
        ----------
        idadb : IdaDataBase
            the database connector
        """

        self.idadb = idadb

    def list_models(self) -> IdaDataFrame:
        """
        Retrieve existing models (in the current database) and returns result as a data frame.

        Returns
        -------
        IdaDataFrame
            the data frame with the list of models
        """

        return IdaDataFrame(idadb=self.idadb, tablename='INZA.V_NZA_MODELS')

    def model_exists(self, name: str) -> bool:
        """
        Checks if a model with the given name exists.

        Parameters
        ----------
        name : str
            the name of model

        Returns
        -------
        bool
            True if the model with the given name exists, otherwise False
        """

        ret = self.idadb.ida_query(f'call NZA..MODEL_EXISTS(\'model={name}\')')
        return not ret.empty and ret[0]

    def drop_model(self, name: str):
        """
        Drops the model with the given name. Does noting if there is no such madel in the database.

        Parameters
        ----------
        name : str
            the name of model
        """

        if self.model_exists(name):
            self.idadb.ida_query(f'call NZA..DROP_MODEL(\'model={name}\')')
    
    def copy_model(self, name: str, copy_name: str):
        """
        Duplicates the given analytics model to a new model with the given new name. 
        The source model can be in another database.

        Parameters
        ----------
        name : str
            the model to be duplicated. It can be qualified by a database name (<database>..<model>)
            if not in the current database.
        
        copy_name : str
            The name of the copy of the model to be created in the current database
        """
        
        self.idadb.ida_query(f'call NZA..COPY_MODEL(\'model={name},copy={copy_name}\')')
    

    def alter_model(self, model: str, name: str=None, owner: str=None, description: str=None, 
                    copyright: str=None, app_name: str=None, app_version: str=None, category: str=None):
        """
        Alters the properties of the given model

        Parameters
        ----------
        model : str
            the model to be altered
        name : str, optional
            the new name of the model. If not specified, the model is not renamed. 
            Otherwise, the model name and the names of the managed tables are changed
        owner : str, optional
            the new owner of the model. If not specified, the owner of the model does not change. 
            Otherwise, the new owner will get all model privileges.
        description : str, optional
            the description for the model. If not specified, the description of the model does not change.
            Otherwise, the given description will be associated to the model
        copyright : str, optional
            the copyright statement for the model, used when exporting the model to PMML. 
            If not specified, the copyright statement of the model does not change. 
            Otherwise, the given copyright statement will be associated to the model.
        app_name : str, optional
            the application that created the model, used when exporting the model to PMML. 
            If not specified, the application that created the model does not change. 
            Otherwise, the given application name will be associated to the model
        app_version : str, optional
            the version of the application that created the model, used when exporting the model to PMML. 
            If not specified, the application version of the model does not change. 
            Otherwise, the given application version will be associated to the model
        category : str, optional
            the user-defined category for the model. If not specified, the category of the model does not change. 
            Otherwise, the given category will be associated to the mode
        """
        params = {
            'model': model,
            'name': name,
            'owner': owner,
            'description': description,
            'copyright': copyright,
            'appname': app_name,
            'appversion': app_version,
            'category': category
        }
        params_s = map_to_props(params)
        if self.model_exists(model):
            self.idadb.ida_query(f'call NZA..ALTER_MODEL(\'{params_s}\')')
        
    def grant_model(self, model: str, privilege: str, user: List[str]=None, 
                        group: List[str]=None, grant_option: bool=False):
        """
        Grants one or more privileges on an analytics model to users and/or groups

        Parameters
        ----------
        model : str
            the model to be granted privileges
        privelige : List[str]
            list of privileges. Allowed privileges are: list, select, alter, update, drop
        user : List[str], optional
            list of users to grant privileges to. If not specified, no user is granted 
            the privileges, and the parameter group has to be specified
        group : List[str], optional
            list of groups to grant privileges to. If not specified, no group is granted 
            the privileges, and the parameter user has to be specified
        grant_option : bool, optional
            flag indicating if the users or groups can further grant the given privileges 
            to other users or groups. Default value - False
        """
        params = {
            'model': model,
            'privilege': privilege,
            'user': user,
            'group': group,
            'grantoption': grant_option
        }
        params_s = map_to_props(params)
        if self.model_exists(model):
            self.idadb.ida_query(f'call NZA..GRANT_MODEL(\'{params_s}\')')

    def revoke_model(self, model: str, privilege: List[str], user: List[str]=None, group: List[str]=None):
        """
        Revokes one or more privileges on an analytics model from users and/or groups

        Parameters
        ----------
        model : str
            the model to revoke privileges
        privelige : List[str]
            list of privileges. Allowed privileges are: list, select, alter, update, drop
        user : List[str], optional
            list of users to revoke privileges to. If not specified, no user privileges
            are revoked, and the parameter group has to be specified
        group : List[str], optional
            list of groups to revoke privileges to. If not specified, no group privileges
            are revoked, and the parameter user has to be specified
        """
        params = {
            'model': model,
            'privilege': privilege,
            'user': user,
            'group': group
        }
        params_s = map_to_props(params)
        if self.model_exists(model):
            self.idadb.ida_query(f'call NZA..REVOKE_MODEL(\'{params_s}\')')

    def list_privileges(self, user: str=None, grant: bool=False):
        """
        Lists the effective privileges of a selected or all users on all models 
        for which you have the LIST privilege

        Parameters
        ----------
        user : str, optional
            the user whose privileges are listed. If not specified, the privileges of all users are 
            listed
        grant : bool, optional
            if true, the grant permissions are displayed. If false (default), 
            the regular object permissions are displayed
        """
        params = {
            'user': user,
            'grant': grant
        }
        cursor = self.idadb._con.cursor()
        params_s = map_to_props(params)
        cursor.execute(f'call NZA..LIST_PRIVILEGES(\'{params_s}\')')
        return ' ' + cursor.notices[0]
        # return self.idadb.ida_query(f'call NZA..LIST_PRIVILEGES(\'{params_s}\')')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.decision_trees import DecisionTreeClassifier
from nzpyida.analytics.tests.conftest import MOD_NAME, TAB_NAME_TRAIN, df_train

import pytest

MOD_NAME_COPY = MOD_NAME+'_COPY'
PRIVILEGES_OUTPUT1 = "INZAUSER  | TEST_MOD1  |   X X   X"""
PRIVILEGES_OUTPUT2 = "INZAUSER  | TEST_MOD1  | X X X   X"
PRIVILEGES_OUTPUT3 = "INZAUSER  | TEST_MOD1  |   X X   X"

@pytest.fixture(scope='module')
def clear_up(idadb: IdaDataBase):
    mm = ModelManager(idadb)
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if mm.model_exists(MOD_NAME_COPY):
        mm.drop_model(MOD_NAME_COPY)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if mm.model_exists(MOD_NAME_COPY):
        mm.drop_model(MOD_NAME_COPY)


def test_model_manager(idadb, clear_up):
    mm = ModelManager(idadb)
    assert mm

    # test listing models
    models_list = mm.list_models()
    assert models_list.exists()
    begin_length = len(mm.list_models()) 

    # verify model does not exist
    assert not mm.model_exists(MOD_NAME)

    # create model and check if mm.model_exits shows it
    idf_train = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True)
    DecisionTreeClassifier(idadb, model_name=MOD_NAME).fit(idf_train, id_column='ID', target_column='B')
    assert mm.model_exists(MOD_NAME)

    # test copy model
    mm.copy_model(name=MOD_NAME, copy_name=MOD_NAME_COPY)
    assert mm.model_exists(MOD_NAME_COPY)

    # test listing models 2
    assert len(mm.list_models()) == 2 + begin_length

    # test droping model
    mm.drop_model(MOD_NAME)
    assert not mm.model_exists(MOD_NAME)

    # test alter model
    mm.alter_model(MOD_NAME_COPY, name=MOD_NAME, owner="INZAUSER", description="DecTree model",
                   copyright="Copyright (c) 2023. IBM Corp. All rights reserved.", category="DecTree")
    
    assert mm.model_exists(MOD_NAME)

    lm_ida = mm.list_models()
    lm = lm_ida.as_dataframe()
    lm = lm[lm["MODELNAME"]==MOD_NAME]
    assert len(lm) == 1
    assert lm["OWNER"].iloc[0] == "INZAUSER"
    assert lm["DESCRIPTION"].iloc[0] == "DecTree model"
    assert lm["COPYRIGHT"].iloc[0] == "Copyright (c) 2023. IBM Corp. All rights reserved."
    assert lm["USERCATEGORY"].iloc[0] == "DecTree"
    assert lm["CREATOR"].iloc[0] == "ADMIN"
    assert lm["ALGORITHM"].iloc[0] == "Decision Tree"

    # test grant privileges

    mm.revoke_model(MOD_NAME, ["list", "update"], user=["INZAUSER"])

    assert PRIVILEGES_OUTPUT1 in mm.list_privileges()

    mm.grant_model(MOD_NAME, privilege=["list"], user=["INZAUSER"])

    assert PRIVILEGES_OUTPUT2 in mm.list_privileges()

    assert PRIVILEGES_OUTPUT3 in mm.list_privileges(grant=True)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""

Many types of business-relevant or scientific data have values that change over time. 
Some typical examples are:
- Daily sales figures for a store
- Energy consumption readings from household electric meters 
- Price per gallon at a local gas station
It is often useful to analyze the behavior of such changes, both to describe the development 
over time, specifically for a particular trend and seasonality, as well as to predict unknown 
values of the series, usually for the future. A typical area of application is supply chain
management, where future needs may be predicted based on past trends.

A time series is a sequence of numerical data values, measured at successive, but not necessarily 
equidistant—points in time. Examples are daily stock prices, monthly unemployment counts, 
or annual changes in global temperature. The two main goals of time series analysis are to 
understand the underlying patterns which are represented by the observed data and to make forecasts.
"""
from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling
from nzpyida.analytics.utils import call_proc_df_in_out
from nzpyida.analytics.model_manager import ModelManager

class TimeSeries(PredictiveModeling):
    """
    Time Series Model
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Time Series
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "TIMESERIES"
        self.has_print_proc = True
    
    def fit_predict(self, in_df: IdaDataFrame, time_column: str, target_column: str, by_column: str=None,
            out_table: str=None, description_table: str=None, algorithm: str='ExponentialSmoothing', 
            interpolation_method: str='linear', from_time=None, to_time=None, forecast_horizon: str=None,
            forecast_times: str=None, trend: str=None, seasonality: str=None, period: float=None, 
            unit: str=None, p: int=None, d: int=None, q: int=None, sp: int=None, sd: int=None, sq: int=None, 
            saesonally_adjusted_table: str=None ) -> IdaDataFrame:
        """
        Predicts future values of series of timed numeric values

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame
        
        time_column : str
            the input data frame column which define an order on the numeric values
        
        target_columns : str
            the input data frame column which contains the numeric values

        by_column : str
            the input data frame column which uniquely identifies a serie of values.
            If not specified, all numeric values belong to only one time series.
        
        out_table : str
            the output data frmae containing predicted future values. This parameter 
            is not allowed for algorithm = SpectralAnalysis. If not specified, 
            no output table is written out
        
        description_table : str
            the optional input data frame containing the name and descriptions of the 
            time series. The table must contain following columns: <by_column>, 'NAME'=str, 
            'DESCRIPTION'=str. If not specified, the series do not have a name or a description
        
        algorithm : str
            the time series algorithm to use. Allowed values are: ExponentialSmoothing, 
            ARIMA, SeasonalTrendDecomposition, SpectralAnalysis

        interpolation_method : str
            the interpolation method. Allowed values are: linear, cubicspline, exponentialspline

        from_time : same as type of <time column>
            the value of column time to start the analysis from. If not specified, the analysis 
            starts from the first value of the time series in the input table

        to_time : same as type of <time column>
            the value of column time to stop the analysis at. If not specified, the analysis 
            stops at the last value of the time series in the input table

        forecast_horizon : str
            the value of column time until which to predict. This parameter is not allowed for 
            algorithm=SpectralAnalysis. If not specified, the algorithm determines itself 
            until which time it predicts values

        forecast_times : str
            list of semi-column separated values of column time to predict at. This parameter 
            is not allowed for algorithm=SpectralAnalysis. If not specified, the times to predict 
            values at is determined by the algorithm

        trend : str
            the trend type for algorithm=ExponentialSmoothing. Allowed values are: N (none), 
            A (addditive), DA (damped additive), M (multiplicative), DM (damped multiplicative). 
            If not specified, the trend type is determined by the algorithm

        seasonality : str
            the seasonality type for algorithm=ExponentialSmoothing. Allowed values are: N (none), 
            A (addditive), M (multiplicative). If not specified, the seasonality type is 
            determined by the algorithm

        period : float
            the seasonality period. This parameter is not allowed for algorithm=SpectralAnalysis. 
            If not specified, the seasonality period is determined by the algorithm. If set to 0, 
            no seasonality period will be considered by the algorithm
        
        unit : str
            the seasonality period unit. This parameter is not allowed for algorithm=SpectralAnalysis. 
            This parameter must be specified if the parameter period is specified and the <time_column>  
            is of type date, time or timestamp. Otherwise, it must not be spe- cified. Allowed values are: 
            ms, s, min, h, d, wk, qtr, q, a, y

        p : int
            the parameter p for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically

        d : int
            the parameter d for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically
        
        q : int
            the parameter q for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically

        sp : int
            the seasonal parameter SP for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically

        sd : int
            the seasonal parameter SD for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically

        sq : int
            the seasonal parameter SQ for algorithm=ARIMA, either equal to or below specified value. 
            If not specified, the algorithm will determine its best value automatically

        saesonally_adjusted_table : str
            the output table containing seasonally adjusted values. This parameter is not allowed 
            for algorithm=SpectralAnalysis or algorithm=ARIMA. If not specified, no output table 
            is written out
        """

        params = {
            'model': self.model_name,
            'time': time_column,
            'target': target_column,
            'by': by_column,
            'desctable': description_table,
            'algorithm': algorithm,
            'interpolationmethod': interpolation_method,
            'from': from_time,
            'to': to_time,
            'forecasthorizon': forecast_horizon,
            'forecasttimes': forecast_times,
            'trend': trend,
            'seasonality': seasonality,
            'period': period,
            'unit': unit,
            'p': p,
            'd': d,
            'q': q,
            'SP': sp,
            'SD': sd,
            'SQ': sq,
            'seasadjtable': saesonally_adjusted_table,
        }

        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        ModelManager(self.idadb).drop_model(self.model_name)

        return call_proc_df_in_out(proc=self.fit_proc, in_df=in_df, params=params,
            out_table=out_table)[0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
This module consists of algorithms used to describe the empirical distribution
of single attributes or the joint distribution of multiple—usually two—attributes.
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.utils import call_proc_df_in_out


def moments(in_df: IdaDataFrame, in_column: str, by_column: str=None,
    out_table: str=None) -> IdaDataFrame:
    """
    Moments are quantities used to describe certain aspects of continuous attribute
    distributions. Of particular interest are the central moments or moments around
    the mean.

    This function Calculates the moments of a numeric input column: mean, variance,
    stand-ard deviation, skewness and (excess) kurtosis as well as the
    count of cases, the minimum and the maximum.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table column

    by_culumn : str, optional
        the input table column which splits the data into groups for which the
        operation is to be per-formed

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc='MOMENTS', in_df=in_df, params=params, out_table=out_table)[0]

def quantile(in_df: IdaDataFrame, in_column: str, quantiles: List[int],
    out_table: str=None) -> IdaDataFrame:
    """
    Quantiles constitute a convenient and intuitive description of continuous attribute
    distribution that allow observation of location, dispersion, and asymmetry.
    Quantiles of a continuous attribute are values from its range taken at regular intervals
    of its cumulative distribution.

    This function calculates quantile limit(s) for a numeric column.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table column

    quantiles : List[int]
        the list of quantiles to be calculated.
        Quantiles are values between 0 and 1 indicating the percentage of sorted
        values to be considered in each quantile.

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column,
        'quantiles': quantiles
    }
    return call_proc_df_in_out(proc='QUANTILE', in_df=in_df, params=params, out_table=out_table)[0]

def outliers(in_df: IdaDataFrame, in_column: str, multiplier: float=1.5,
    out_table: str=None) -> IdaDataFrame:
    """
    Outliers are the values below the first quartile or above the third quartile by more than
    the inter-quartile range multiplied by a coefficient of the attribute, which controls
    the aggressiveness of outlier detection.

    This function detects outliers of a numeric attribute (a column).

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table column

    multiplier : float, optional
        the value of the IQR multiplier

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column,
        'multiplier': multiplier
    }
    return call_proc_df_in_out(proc='OUTLIERS', in_df=in_df, params=params, out_table=out_table)[0]

def unitable(in_df: IdaDataFrame, in_column: str, out_table: str=None) -> IdaDataFrame:
    """
    A univariate frequency table describes the distribution of a discrete attribute
    by providing the occurrence count for each unique value.

    This function creates a univariate frequency table for one column of the input table.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table column

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column
    }
    return call_proc_df_in_out(proc='UNITABLE', in_df=in_df, params=params, out_table=out_table)[0]

def bitable(in_df: IdaDataFrame, in_column: List[str], freq: bool=False, cum: bool=False,
    out_table: str=None) -> IdaDataFrame:
    """
    A bivariate frequency table describes the joint probability distribution of two
    discrete attributes, by providing the occurrence count for each distinct combination
    of their values.

    This function creates a bivariate frequency table for two columns of the input table.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the list of numeric input table columns, must be two (if there are more, they
        will be ignored)

    freq : bool, optional
        flag indicating whether frequencies should be attached to the output table

    cum : bool, optional
        flag indicating whether cumulative frequencies should be attached to the output 
        table (setting this flag automatically sets freq flag as frequencies have to be 
        calculated prior to cumulative frequencies)

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column,
        'freq': freq,
        'cum': cum
    }
    return call_proc_df_in_out(proc='BITABLE', in_df=in_df, params=params, out_table=out_table)[0]

def histogram(in_df: IdaDataFrame, in_column: str, nbreaks: int=None, right: bool=True,
    btable: str=None, bcolumn: str=None, density: bool=False, midpoints: bool=False,
    freq: bool=False, cum: bool=False, out_table: str=None) -> IdaDataFrame:
    """
    A histogram is a frequency table counterpart for continuous attributes. Although usually
    presented visually as a graph, it can be considered a table providing occurrence counts
    for a series of disjoint intervals covering the range of the attribute. The intervals
    can be of equal or inequal width. The number of intervals and their boundaries can be
    specified manually to ensure the histogram is most meaningful and readable, or adjusted
    automatically to the distribution.

    This function creates histograms. The number of bins and the bins themselves
    can be specified or are automatically calculated.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the input table column to build the histogram onto

    nbreaks : int, optional
        the number of bins for the histogram.
        If not specified, the number of bins is calculated auto-matically.

    right : bool, optional
        the flag indicating whether the histogram bins should be
        right-closed (true) or right-open (false)

    btable : str, optional
        the input table with breaks for the histogram.
        If not specified, the bins are calculated automat-ically, using the
        parameter <nbreaks> if specified.

    bcolumn : str, optional
        the <btable> column containing the breaks for the histogram.
        This column must be specified if the parameter <btable> is specified.

    density : bool, optional
        flag indicating whether densities should be attached to the output table

    midpoints : bool, optional
        flag indicating whether the midpoints of the bins should be attached to the output table

    freq : bool, optional
        flag indicating whether frequencies should be attached to the output table

    cum : bool, optional
        flag indicating whether cumulative frequencies should be attached to the output table 
        (setting this flag automatically sets freq flag as frequencies have to be calculated 
        prior to cumulative frequencies)

    out_table : str, optional
        the output table to write the moments into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """

    params = {
        'incolumn': in_column,
        'nbreaks': nbreaks,
        'right': right,
        'btable': btable,
        'bcolumn': bcolumn,
        'density': density,
        'midpoints': midpoints,
        'freq': freq,
        'cum': cum
    }
    return call_proc_df_in_out(proc='HIST', in_df=in_df, params=params, out_table=out_table)[0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The divisive clustering algorithm is a computationally efficient, top-down approach 
to creating hierarchical clustering models. Conceptually, it can be thought of as 
a wrapper around the k -means algorithm (with a specialized method for initial centroid setting), 
running the algorithm several times to divide clusters into subclusters. 
The internal k-means algorithm assumes a fixed k =2 value.

The divisive clustering algorithm may return different results for the same data set 
and the same random generator seed when you use different input data distribution or 
a different number of dataslices. This is due to the behavior of the random number 
generator, which generates random sequences depending on the number of dataslices 
and data distribution. The algorithm returns the same model when you use the same machine, 
the same input data distribution, and the same random seed.

The cluster formation process of the divisive clustering algorithm begins with a single cluster 
containing all training instances, then the first invocation of k-means divides it into two 
subclusters by creating two descendant nodes of the clustering tree. Subsequent invocations 
divide these clusters into more subclusters, and so on, until a stop criterion is satisfied. 
Stop criterion can be specified by the maximum clustering tree depth or by the minimum required 
umber of instances for further partitioning. The resulting hierarchical clustering tree can be 
used to classify instances by propagating them down from the root node, and choosing at each 
level the best matching sub-cluster with respect to the instance’s distance from sub-cluster centers.

The internal k-means process of the divisive clustering algorithm operates using the ordinary 
k-means algorithm (with the modified initial centroid generation), discussed in the K-Means 
Clustering section, using a fixed value of k=2 and working with the subset of data from the 
parent cluster. The initial centroid generation consists two steps: random generation n>>k 
candidates and then selection of outermost pair of candidates. The cluster center representation 
and distance measures remain the same. The numbering scheme for clusters in a clustering tree is 
the same as decision trees: the root node is number 1, and the descendants of node number 'i' have 
numbers '2i' and '2i+1' 

Additionally, leaves, which are clusters with no subclusters, are designated by negative numbers.
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, make_temp_table_name
from nzpyida.analytics.utils import get_auto_delete_context
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling

class BisectingKMeans(PredictiveModeling):
    """
    Divisive Clustering
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the clusterer class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'DIVCLUSTER'
        self.predict_proc = 'PREDICT_DIVCLUSTER'
        self.score_proc = 'MSE'
        self.target_column_in_output = 'CLUSTER_ID'
        self.id_column_in_output = 'ID'
        self.has_print_proc = True

    def fit(self, in_df: IdaDataFrame, id_column: str=None, target_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, out_table: str=None, distance: str='euclidean',
        max_iter: int=5, min_split: int=5, max_depth: int=3, rand_seed: int=12345) -> IdaDataFrame:
        """
        Builds a Hierarchical Clustering model using a divisive method (top-down). 
        The K- means algorithm is used recursively. The hierarchy of clusters is 
        represented in a binary tree structure (each parent node has exactly 2 children node). 
        The leafs of the cluster tree are identified by negative numbers. 
        The divisive clustering algorithm may return different results for the same dataset 
        and the same random generator seed when you use different input data distribution 
        or a different number of dataslices. This is due to the behavior of the random number 
        generator, which generates random sequences depending on the number of dataslices 
        and data distribution. The algorithm returns the same model when the same ma-chine, 
        the same input data distribution, and the same random seed is used.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id
        
        target_column : str, optional
            the input table column representing a class or a value to predict, 
            this column is ignored by the Hierarchical Clustering algorithm

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        out_table : str, optional
            the output table where clusters are assigned to each input table record

        distance : str, optional
            the distance function. Allowed values are: euclidean, norm_euclidean, manhattan,
            canberra, maximum, mahalanobis.

        max_iter : int, optional
            the maximum number of iterations to perform in the base K-means Clustering algorithm
        
        min_split : int, optional
            the minimum number of instances per cluster that can be split
        
        max_depth : int, optional
            the maximum number of cluster levels (including leaves)

        rand_seed : int, optional
            the random generator seed
        
        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers, cluster_id and distance to cluster center
        """

        auto_delete_context = None
        if not out_table:
            auto_delete_context = get_auto_delete_context('out_table')
            out_table = make_temp_table_name()

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'distance': distance,
            'maxiter': max_iter,
            'minsplit': min_split,
            'maxdepth': max_depth,
            'randseed': rand_seed,
            'outtable': out_table
        }

        self._fit(in_df=in_df, params=params)

        if auto_delete_context:
            auto_delete_context.add_table_to_delete(out_table)

        return IdaDataFrame(self.idadb, out_table)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None, level: int=-1) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the assigned clusters will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model
        
        level : int, optional
            the level of the cluster hierarchy which should be applied to the data. 
            For level=-1, only the leaves of the clustering tree are considered

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column,
            'level': level
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str,
        id_column: str=None, level: int=-1) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id
        
        level : int, optional
            the level of the cluster hierarchy which should be applied to the data. 
            For level=-1, only the leaves of the clustering tree are considered

        Returns
        -------
        float
            the model score
        """

        params = {
            'id': id_column,
            'level': level
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
This module consists of algorithms used to detect and quantify relationships between attributes.
"""

from nzpyida.analytics.utils import call_proc_df_in_out, make_temp_table_name, out_str_to_df
from nzpyida.frame import IdaDataFrame
from typing import List
import pandas as pd


def corr(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This stored procedure calculates the correlation between two numeric input columns, 
    either in the whole input table or within the groups defined in the column specified by parameter <by>. 
    Correlation is a measure saying how easily it is to predict one column from the other. 
    It takes a value between -1 (inversely correlated) and 1 (correlated), 
    0 means that the two columns are independent.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two numeric input table columns, separated by a semicolon (;). 
        Optionally, a third numeric column can be specified for weights followed by :objweight.
    
    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 
        If specified, an output table must be indicated.
    
    out_table : str, optional
        the output table to write the correlations into. 
        This parameter must be specified if parameter by is specified.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    out_df = call_proc_df_in_out(proc="CORR", in_df=in_df, params=params, out_table=out_table)[0]
    temp_table = make_temp_table_name()
    query = f"""create table {temp_table} as
            select case when CORRELATION is null then 0 else CORRELATION
            end as correlation {', ' + by_column if by_column else ''}
            from {out_table};
            drop table {out_table};
            alter table {temp_table} rename to {out_table};
            """
    in_df._idadb.ida_query(query)
    return out_df


def cov(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This function calculates the covariance of two numeric input columns, 
    either in the whole input table or within the groups defined in the column specified by parameter <by>. 
    Covariance is a measure say-ing how easily it is to predict one column from the other. 
    It takes a negative value when both columns are inversely correlated, 
    a positive value when they are correlated, 0 means that the two columns are inde-pendent.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two numeric input table column, separated by a semicolon

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 
        If specified, an output table must be indicated.
    
    out_table : str, optional
        the output table to write the covariances into. 
        This parameter must be specified if parameter <by_column> is specified.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc="COV", in_df=in_df, params=params, out_table=out_table)[0]


def covariance_matrix(in_df: IdaDataFrame, in_column: List[str], out_table: str=None, by_column: str=None):
    """
    This function calculates the matrix of covariances between 
    pairs of numeric input columns divided into two sets X and Y, either in the whole 
    input table or within the groups defined in the column specified by parameter <by_column>. 
    Covariance is a measure saying how easily it is to predict one column from the other. 
    It takes a negative value when both columns are inversely correlated, 
    a positive value when they are correlated, 0 means that the two columns are independent.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str], optional
        the numeric input table columns to calculate covariances for, separated by a semi- colon (;). 
        Each column is followed by :X or :Y to indicate it belongs to set X or set Y. 
        If neither :X nor :Y is specified for any of the input columns, 
        the matrix contains the cov-ariances between all pairs of input columns.

    out_table : str, optional
        the output table to write the covariances

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc="COVARIANCEMATRIX", in_df=in_df, params=params, out_table=out_table)[0]


def spearman_corr(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This function calculates the Spearman rank correlation 
    between two columns whose values are ordered in their respective domain, either in the whole 
    input table or within the groups defined in the column specified by parameter <by_column>. 
    If both columns are of type double, or int, or date, or time, the order is obviously guaranteed. 
    In case of character columns, it is assumed that the order of their values is lexicographic. 
    The Spearman rank correlation is a non-parametric measure of dependence between two variables. 
    It takes a value between -1 (inversely correlated) and 1 (correlated), 
    0 means that the two columns are independent.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two ordinal input table columns, separated by a semicolon (;). 
        Optionally, a third numeric column can be specified for weights followed by :objweight.
    
    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 
        If specified, an output table must be indicated.
    
    out_table : str, optional
        the output table to write the Spearman rank correlations into. 
        This parameter must be specified if parameter <by_column> is specified.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    temp_table = make_temp_table_name()
    if by_column:
        out_df =  call_proc_df_in_out(proc="SPEARMAN_CORR", in_df=in_df, params=params, out_table=out_table)[0]
        query = f"""create table {temp_table} as
            select {by_column + ', '} case when RHO is null then 0 else RHO
            end as RHO, N
            from {out_table};
            drop table {out_table};
            alter table {temp_table} rename to {out_table};
            """
    else:
        output = call_proc_df_in_out(proc="SPEARMAN_CORR", in_df=in_df, params=params, out_table=out_table)[1]
        out_df = in_df._idadb.as_idadataframe(pd.DataFrame(output), tablename=out_table)
        query = f"""create table {temp_table} as
            select case when SPEARMAN_CORR is null then 0 else SPEARMAN_CORR
            end as SPEARMAN_CORR
            from {out_table};
            drop table {out_table};
            alter table {temp_table} rename to {out_table};
            """
    in_df._idadb.ida_query(query)
    return out_df


def mutual_info(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This function calculates the mutual information of two input columns, 
    either in the whole input table or within the groups defined in the column specified by parameter <by>. 
    Mutual Information is a measure saying how easily it is to predict the value of one column 
    from the value of the other column. It takes a positive value, the lower the better predictability.
    
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two input table columns separated by a semicolon
    
    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 
        If specified, an output table must be indicated.
    
    out_table : str, optional
        the output table to write the mutual information into. 
        This parameter must be specified if parameter <by_column> is specified.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc="MUTUALINFO", in_df=in_df, params=params, out_table=out_table)[0]


def chisq(in_df: IdaDataFrame, in_column: List[str], out_table: str=None, by_column: str=None):
    """
    This function calculates the Chi-square value between two input columns, either 
    in the whole input table or within the groups defined in the column specified by parameter <by_column>. 
    Chi-square is a measure saying how easily it is to predict one column from the other. 
    The function then returns probability that the two columns are independent.

    Additionally to the Chi-square statistics, this stored procedure returns the degree of freedom 
    of the input variables and a percentage between 0 and 1.
    If the percentage ranges from 0 to 0.05, the columns are definitely independent,
    If the percentage ranges between 0.05 and 0.95, they are to be treated as independent, though one is not sure,
    If the percentage ranges between 0.95 and 1, then a predictive mutual dependency is definitely there.
        
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two input table columns separated by a semicolon
    
    out_table : str, optional
        the output table to write the Chi-square values

    by_column : str, optional
        he input table column which splits the data into groups for which the operation is to be performed. 
        If specified, an output table must be indicated.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc="CHISQ_TEST", in_df=in_df, params=params, out_table=out_table)[0]


def t_me_test(in_df: IdaDataFrame, in_column: str, mean_value: float, by_column: str=None, out_table: str=None):
    """
    This function calculates the t-Student statistics of an input column with the expected mean, 
    either in the whole input table or within the groups defined in the column specified by parameter <by>. 
    This t-Student test is a measure saying whether or not the column has the given mean.
    
    If the percentage ranges from 0 to 0.05, the expected mean value is significantly too high,
    If the percentage ranges between 0.05 and 0.95, the expected mean value matches the mean value of the column,
    If the percentage ranges between 0.95 and 1, the expected mean value is significantly too low.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table
    
    mean_value : float
        the expected mean value of the input column

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed.
    
    out_table : str, optional
        the output table to write the t-Student statistics

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'mean': mean_value
    }
    return call_proc_df_in_out(proc="T_ME_TEST", in_df=in_df, params=params, out_table=out_table)[0]


def t_umd_test(in_df: IdaDataFrame, in_column: str, class_column: str, out_table: str=None, by_column: str=None):
    """
    This function calculates the t-Student statistics of a column whose values are split into two classes, 
    either in the whole input table or within the groups defined in the column specified by parameter <by>. 
    This t-Student test is a measure saying whether or not the two classes have the same mean value.
    
    If the percentage ranges from 0 to 0.05, the second class has a significantly bigger mean value than the first class,
    If the percentage ranges between 0.05 and 0.95, the two classes have the same mean value,
    If the percentage ranges between 0.95 and 1, the second class has a significantly smaller mean value than the first class.
    
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the numeric input table
    
    class_column : str
        the input table column which splits data into two classes. 
        The class column name is followed by two class values preceded by a colon (:), eg. 'column_name:class_1:sclass_2'

    out_table : str, optional
        the output table to write the t-Student statistics

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'class': class_column
    }
    return call_proc_df_in_out(proc="T_UMD_TEST", in_df=in_df, params=params, out_table=out_table)[0]

def t_pmd_test(in_df: IdaDataFrame, in_column: List[str], expected_diff: float, out_table: str=None, by_column: str=None):
    """
    This function calculates the t-Student statistics of two paired columns, either in the whole 
    input table or within the groups defined in the column specified by parameter <by>. 
    This t-Student test is a measure saying whether or not the two columns take different values, 
    on average distinct by <expected_diff>
    
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two numeric input table columns, separated by a semicolon (;). 
        One column must be followed by :X, the other column by :Y.
    
    expected_diff : float
        the expected difference between the mean values of the input columns (Y - X)

    out_table : str, optional
        the output table to write the t-Student statistics

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed.

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'expdiff': expected_diff
    }
    return call_proc_df_in_out(proc="T_PMD_TEST", in_df=in_df, params=params, out_table=out_table)[0]


def t_ls_test(in_df: IdaDataFrame, in_column: List[str], slope: float, by_column: str=None, out_table: str=None):
    """
    This function calculates the t-Student statistics between two input columns, either in the whole 
    input table or within the groups defined in the column specified by parameter <by>. This t-Student test 
    is a measure saying whether or not the two columns are linearly related with a given slope.
    
    If the percentage ranges from 0 to 0.05, the slope is too high,
    If the percentage ranges between 0.05 and 0.95, the indicated slope is acceptable, 
    the lower the percentage the more independent the columns are,
    If the percentage ranges between 0.95 and 1, the slope is too low.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two numeric input table columns, separated by a semicolon (;). 
        One column must be followed by :X, the other column by :Y.
    
    slope : float
        the expected direction of dependence

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed.
    
    out_table : str, optional
        the output table to write the t-Student statistics

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'slope': slope
    }
    return call_proc_df_in_out(proc="T_LS_TEST", in_df=in_df, params=params, out_table=out_table)[0]

#error in documentation
def mww_test(in_df: IdaDataFrame, in_column: str, class_column: str, by_column: str=None, out_table: str=None):
    """
    This function executes the Mann-Whitney-Wilcoxon test on a column whose values are split into 
    two classes, either in the whole input table or within the groups defined in the column specified 
    by parameter <by>. The Mann-Whitney-Wilcoxon test is a statistical hypothesis test that determines 
    whether one of two samples of independent observations tends to have larger values than the other. 
    This is indicated by the component of the output called pp. If pp < 0.05, then one of the two classes 
    tends to have larger values than the other. The class which tends to have lower-ranked values is indicated.
    
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the input table column. It does not need to be numerical, but must be ordered in its domain
    
    class_column : str
        the input table column which splits data into two classes. 

    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed.

    out_table : str, optional
        the output table to write the MWW statistics into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'class': class_column
    }
    if by_column:
        return call_proc_df_in_out(proc="MWW_TEST", in_df=in_df, params=params, out_table=out_table)[0]
    else:
        output_string = call_proc_df_in_out(proc="MWW_TEST", in_df=in_df, params=params, out_table=out_table)[1][0]
        return in_df._idadb.as_idadataframe(out_str_to_df(str(output_string)), tablename=out_table)


def wilcoxon_test(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This function executes the Wilcoxon test on two numeric columns, either in the whole input 
    table or within the groups defined in the column specified by parameter <by>. The Wilcoxon test 
    compares in non-parametric manner two variables to state if they are different in their mean value. 
    This is indicated by the component of the output called pp. If pp < 0.05, then one of the two columns 
    tends to have larger values than the other. The class which tends to have lower-ranked values is indicated

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the two numeric input table columns, separated by a semicolon (;)
    
    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 

    out_table : str, optional
        the output table to write the Wilcoxon statistics into

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    if by_column:
        return call_proc_df_in_out(proc="WILCOXON_TEST", in_df=in_df, params=params, out_table=out_table)[0]
    else:
        output_string = call_proc_df_in_out(proc="WILCOXON_TEST", in_df=in_df, params=params, out_table=out_table)[1][0]
        return in_df._idadb.as_idadataframe(out_str_to_df(str(output_string)), tablename=out_table)


def canonical_corr(in_df: IdaDataFrame, in_column: List[str], by_column: str=None, out_table: str=None):
    """
    This function calculates the canonical correlation of two sets of input columns, either in the whole 
    input table or within the groups defined in the column specified by parameter <by>. 
    Canonical correlation is a measure saying how easily it is to predict one column set from the other. 
    It takes a value between -1 (in- versely correlated) and 1 (correlated), 
    0 means that the two sets of columns are independent.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str] 
        the numeric input table columns to be correlated, separated by a semicolon (;). 
        Each column is followed by :X or :Y to indicate it belongs to set X or set Y
    
    by_column : str, optional
        the input table column which splits the data into groups for which the operation is to be performed. 

    out_table : str, optional
        the output table to write the correlations into. 

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc="CANONICAL_CORR", in_df=in_df, params=params, out_table=out_table)[0]


def anova_crd_test(in_df: IdaDataFrame, in_column: List[str], treatment: str, by_column: str=None, out_table: str=None):
    """
    This function analyzes the variance of one or several observations for different treatments. 
    It assumes that the input table contains one or several columns with numerical (double) 
    observation results of an experiment concerning treatments indicated by the treatment parameter.

    The One-way ANOVA considers independent samples (the treatments) while the Completely Randomized Design 
    considers equally sized "samples". The implementation covers both cases. For more information, 
    see 'Completely Randomized Design'.
    
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the input table observation columns, separated by a semi-colon (;) 
    
    treatment : str
        the input table column identifying a unique treatment

    by_column : str, optional
        the input table column which uniquely identifies a group on which to perform ANOVA
    
    out_table : str, optional
        the output table

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'treatment': treatment
    }
    out_df = call_proc_df_in_out(proc="ANOVA_CRD_TEST", in_df=in_df, params=params, out_table=out_table)[0]
    temp_table = make_temp_table_name()
    usual_columns = ', '.join(['TOTNO', 'TOTSU', 'TOTMEAN', 'TOTSS', 'SSCTOT', 'SSCBETWEEN', 
                                      'DFBETWEEN', 'SSCWITHIN', 'DFWITHIN', 'F'])
    query = f"""create table {temp_table} as
            select {'INCOLUMN, ' if len(in_column)>1 else ''} 
            {by_column + ', ' if by_column else ''} 
            {usual_columns},
            case when p is null then 0 else p
            end as p 
            from {out_table};
            drop table {out_table};
            alter table {temp_table} rename to {out_table};
            """
    in_df._idadb.ida_query(query)
    return out_df


def anova_rbd_test(in_df: IdaDataFrame, in_column: str, treatment: str, block: str, by_column: str=None, out_table: str=None):
    """
    This function analyzes the variance of one or several observations for different blocks of treatments. 
    It assumes that the input table contains one or several columns with numerical (double) observation 
    results of an experiment concerning treatments indicated by the treatment parameter. 
    The treatments are performed repeatedly in various blocks, e.g. from different laboratories 
    where experiments are carried out. This means there are independent samples (treatments) 
    repeatedly drawn for each block. In particular, the number of observations 
    for the same treatment in different blocks should be the same.

    Generally, we would expect differences between blocks (we split into blocks just to reduce 
    the variance resulting) and are curious about the differences concerning treatments.

    The function returns, among others, the p-value of the F test, related to the block split and the group split of the data. 
    If the p-value is larger than 0.95 then it can be said that the hypothesis is incorrect or in this case the treatment or block split had no effect.

    blsscbetween - the sum of squares between the blocks of treatments (around the overal mean), 
    bldfbetween - the number of degrees of freedom between the blocks of treatments, 
    sscwithin - the sum of squares within the treatments (around the mean of each treatment), 
    dfwithin - the sum of the number of degrees of freedom within each treatment, 
    grsscbetween - the sum of squares between the groups (around the overal mean), 
    grdfbetween - the number of degrees of freedom between the groups,
    fbl - the value of the F-statistics for the blocks of treatments,
    pbl - the probability that the true F statistics is lower or equal to the F-statistics above (computed from the sample),
    fgr - the value of the F-statistics for the groups,
    pgr - the probability that the true F statistics is lower or equal to the F-statistics above (computed from the sample).

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the input table observation columns, separated by a semi-colon (;) 
    
    treatment : str
        the input table column identifying a unique treatment

    block : str
        the input table column identifying a unique block of treatments
        
    by_column : str, optional
        the input table column which uniquely identifies a group on which to perform ANOVA
    
    out_table : str, optional
        the output table

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'treatment': treatment,
        'block': block
    }
    return call_proc_df_in_out(proc="ANOVA_RBD_TEST", in_df=in_df, params=params, out_table=out_table)[0]


def manova_one_way_test(in_df: IdaDataFrame, in_column: str, factor1: str, id_column: str=None, by_column: str=None, 
                        table_type: str='trcv', out_table: str=None, timecheck: str=None):
    """
    This function performs one-way analysis of variance/covariance aiming to tell whether or not 
    the groups of data identified by factor1 have the same mean value in all dependent variables or not. 
    As an output 4 matrices (for each task) are produced that are stored in the output table 
    the first matrix describes the ground means of all dependent variables (row vector) 
    the second matrix describes the covariance table stat-istics and their p-values (row vector) 
    the third matrix lists the eigenvalues of the covariance table statistics (column vector) 
    the forth matrix lists the eigenvectors of the covariance table statistics (row of column vec-tors)

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str, optional
        the input table observation columns (dependent variables), separated by a semi-colon (;) 
        used for type='column' only, for type='trcv' these are the values for col-column greater equal 2 
    
    factor1 : str, optional
        the input table column identifying a first factor (so-called treatment in RBD/CRD nomenclature) 
        used for type='column' only, for type='trcv' these are the values for col-column equal 1
    
    id_column : str,  optional
        the input table column which uniquely identifies records used for type='column' only, 
        for type='trcv' not needed due to the structure of the table

    by_column : str, optional
        the input table column which splits the input table into subtables, on each of them a separate MANOVA is run 
        If not specified, the whole input table is subject of a single MANOVA run. 
        used for type='column' only, for type='trcv' it is by default the column named id_task
    
    table_type : str, optional
        the input table form: either 'columns' or 'trcv' trcv stands for "id_task, row, column, value" 
        id_task must be positive integer >= 1. columns means the traditional table representation
    
    out_table : str, optional
        the output table
    
    timecheck : str, optional
        the output will be enriched with execution time of critical sections if set to yes 
        If not specified, the whole input table is subject of a single MANOVA run. 
        used for type='column' only for type='trcv' it is by default the column named id_task

    Returns
    -------
    IdaDataFrame
        the data frame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'factor1': factor1,
        'type': table_type,
        'id': id_column,
        '_timecheck': timecheck
    }
    return call_proc_df_in_out(proc="MANOVA_ONE_WAY_TEST", in_df=in_df, params=params, out_table=out_table)[0]


def manova_two_way_test(in_df: IdaDataFrame, in_column: str, factor1: str, factor2: str, table_type: str, 
                        id_column: str=None, by_column: str=None, out_table: str=None, timecheck: str=None):
    """
    This function performs Two-way analysis of variance/covariance aiming to tell whether or 
    not the groups of data identified by factor1 have the same mean value in all dependent variables or not.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str
        the input table observation columns (dependent variables), separated by a semi-colon (;) 
        used for type='column' only, for type='trcv' these are the values for col-column greater equal 3 
    
    factor1 : str
        the input table column identifying a first factor (so-called treatment in RBD/CRD nomenclature),
        used for type='column' only, for type='trcv' these are the values for col-column equal 1
    
    factor2: str
        the input table column identifying a second factor (so-called block in RBD nomenclature),
        used for type='column' only, for type='trcv' these are the values for col-column equal 2
    
    table_type : str
        the input table form: either 'columns' or 'trcv' trcv stands for "id_task, row, column, value" 
        id_task must be positive integer >= 1. columns means the traditional table representation
    
    id_column : str, optional
        the input table column which uniquely identifies records used for type='column' only,
        for type='trcv' not needed due to the structure of the table

    by_column : str, optional
        the input table column which splits the input table into subtables, on each of them a separate MANOVA is run 
        If not specified, the whole input table is subject of a single MANOVA run. 
        used for type='column' only, for type='trcv' it is by default the column named id_task

    out_table : str, optional
        the output table
    
    timecheck : str, optional
        the output will be enriched with execution time of critical sections if set to yes 
        If not specified, the whole input table is subject of a single MANOVA run. 
        used for type='column' only, for type='trcv' it is by default the column named id_task

    Returns
    -------
    IdaDataFrame
        the data fąrame with requested data
    """
    params = {
        'incolumn': in_column,
        'by': by_column,
        'factor1': factor1,
        'factor2': factor2,
        'type': table_type,
        'id': id_column,
        '_timecheck': timecheck
    }
    return call_proc_df_in_out(proc="MANOVA_TWO_WAY_TEST", in_df=in_df, params=params, out_table=out_table)[0]